//  Libsmtp 0.1 -:-  (C)opyright 2000-2001 Kristian Van Der Vliet
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Library General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Library General Public License for more details.
//
//  You should have received a copy of the GNU Library General Public
//  License along with this library; if not, write to the Free
//  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
//  MA 02111-1307, USA

#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <signal.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include "../inc/libsmtp.h"
#include "../inc/libsmtp_errors.h"

#ifdef __SMTP_DEBUG_
#include <stdio.h>
#endif	/* __SMTP_DEBUG_ */

/* Socket *has* to be a global */
int hGlobalSocket;

/* Timeout must also be global */
unsigned long nSmtpTimeout=30;
unsigned char nSmtpTimedOut=0;	/* 0==FALSE 1==TRUE */

#define __SMTP_ALARM_SET signal(SIGALRM,(void*)smtp_alarm);(void)alarm(nSmtpTimeout);
#define __SMTP_ALARM_CLEAR signal(SIGALRM,SIG_DFL);(void)alarm(0);

int smtp_clear_buf(int nSocket);
void smtp_alarm(void);

int smtp_connect(const char* zServer, int nPort)
{

	struct sockaddr_in sSocketAddr;	/* Socket connection details */
	struct hostent* sServer;			/* Host details */

	char nRecvBuf[3];					/* Receive buffer */
	int nBytesRecv,nBytesSend;		/* Number of bytes read by recv & 
send*/

	char zHostName[255];				/* Hostname of this computer */
	char* nHeloMsg;					/* Message sent when we HELO 
with the server */

	#ifdef __SMTP_DEBUG_
		printf("libsmtp : smtp_connect : Server is \"%s\", port 
%i\n",zServer,nPort);
	#endif /* __SMTP_DEBUG_ */

	/* Lookup the hostname of the server */

	sServer=gethostbyname(zServer);
	if(sServer==NULL)
		return(-ESMTPUNKNOWNSERVER);

	#ifdef __SMTP_DEBUG_
		printf("libsmtp : smtp_connect : IP address %s\n",inet_ntoa(*((struct 
in_addr *)sServer->h_addr)));
	#endif /* __SMTP_DEBUG_ */

	/* Now attempt to open a socket to the server */
	hGlobalSocket=socket(PF_INET,SOCK_STREAM,0);

	sSocketAddr.sin_family=AF_INET;
	sSocketAddr.sin_port=htons(nPort);

	/* FIXME:  Don't even ask.  The code is ugly.  It does it ass	*/
	/* backwards.  It sucks.  But it works.  Want to fix it?  		*/
	/* Please do :) 							
							*/

	if(inet_aton(inet_ntoa(*((struct in_addr 
*)sServer->h_addr)),&(sSocketAddr.sin_addr))==0)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_connect : Connect failed at 
inet_ntoa()\n");
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPCONNECTFAILED);
	}

	memset(&(sSocketAddr.sin_zero),'\0',8);

	/* Set the timeout alarm */
	__SMTP_ALARM_SET

	if(connect(hGlobalSocket, (struct sockaddr*)&sSocketAddr, sizeof(struct 
sockaddr))==-1)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_connect : Connect failed at 
connect()\n");
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPCONNECTFAILED);
	}

	/* Clear the timeout alarm & see if we timedout */
	__SMTP_ALARM_CLEAR
	if(nSmtpTimedOut==1)
		return(-ESMTPTIMEDOUT);

	#ifdef __SMTP_DEBUG_
		printf("libsmtp : smtp_connect : Connected\n");
	#endif	/* __SMTP_DEBUG_ */

	/* We should now have a socket connection to the server,
	on the requested port 
	The SMTP server should send a 220 Service Ready response */

	nBytesRecv=recv(hGlobalSocket,(void*)&nRecvBuf,3,0);
	if(nBytesRecv<3)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_connect : Server did not respond 
correctly!\n");
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPCONNECTFAILED);
	}

	nRecvBuf[3]='\0';		/* Terminate after the first 3 bytes, which 
will be just the return code, as we use strcmp on the buffer */

	if(strcmp(nRecvBuf,"220")!=0)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_connect : Server responded 
%s\n",nRecvBuf);
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPCONNECTFAILED);
	}

	/* Clear out the rest of the buffer */

	smtp_clear_buf(hGlobalSocket);

	/* Create our "HELO" message */
	gethostname(zHostName, sizeof(zHostName));

	nHeloMsg=(char*)malloc(8+strlen(zHostName));
	nHeloMsg=(char*)memset(nHeloMsg,'\0',8+strlen(zHostName));

	nHeloMsg=strcpy(nHeloMsg,"HELO ");
	nHeloMsg=strcat(nHeloMsg,zHostName);
	nHeloMsg=strcat(nHeloMsg,"\r\n\0");

	#ifdef __SMTP_DEBUG_
		printf("libsmtp : smtp_connectt : HELO message is %s\n",nHeloMsg);
	#endif	/* __SMTP_DEBUG_ */

	/* Now send the HELO line & wait for the response */

	#ifdef __SMTP_DEBUG_
		printf("libsmtp : smtp_connect : Sending HELO\n");
	#endif	/* __SMTP_DEBUG_ */

	nBytesSend=send(hGlobalSocket,nHeloMsg,strlen(nHeloMsg),0);

	if(nBytesSend!=strlen(nHeloMsg))
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_connect : Didn't send entire HELO\n");
		#endif	/* __SMTP_DEBUG_ */

	/* Get server response, which should be 250 if we're good to go */
	/* Set the timeout alarm */
	__SMTP_ALARM_SET

	nBytesRecv=recv(hGlobalSocket,(void*)&nRecvBuf,3,0);
	if(nBytesRecv<3)
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_connect : Recieved less than 3 
bytes\n");
		#endif	/* __SMTP_DEBUG_ */

	/* Clear the timeout alarm & see if we timedout */
	__SMTP_ALARM_CLEAR
	if(nSmtpTimedOut==1)
		return(-ESMTPTIMEDOUT);

	nRecvBuf[3]='\0';		/* Terminate after the first 3 bytes, which 
will be just the return code, as we use strcmp on the buffer */

	if(strcmp(nRecvBuf,"250")!=0)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_connect : HELO failed\n");
			printf("libsmtp : smtp_connect : Server responded 
%s\n",nRecvBuf);
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPCONNECTFAILED);
	}

	/* The server is ready to recieve email from us */

	#ifdef __SMTP_DEBUG_
		printf("libsmtp : smtp_connect : Ready to send\n");
	#endif	/* __SMTP_DEBUG_ */

	/* Clear out the rest of the buffer */

	smtp_clear_buf(hGlobalSocket);

	/* The connection is open and the server waiting for us */
	return(EOK);
}

int smtp_sendmail(const char* zTo, const char* zFrom, const char* zBody)
{
	/* Send the mail zBody FROM: zFrom, TO: zTo */

	int nLength;					/* Used to calculate line 
lengths etc */
	int nBytesRecv,nBytesSend;	/* Returns from send() & recv() */
	char nRecvBuf[3];				/* Return code recieve buffer 
*/
	char *nFromMsg, *nRcptMsg;	/* Mail from & rcpt to messages */

	/* Build our "MAIL FROM:" line */
	nLength=strlen(zFrom);
	nLength+=14;	/* 14 is the length of the string "MAIL FROM: " + \r\n\0 */

	nFromMsg=(char*)malloc(nLength);

	nFromMsg=strcpy(nFromMsg,"MAIL FROM: ");
	nFromMsg=strcat(nFromMsg,zFrom);
	nFromMsg=strcat(nFromMsg,"\r\n\0");

	/* Send the MAIL FROM: line */
	nBytesSend=send(hGlobalSocket,nFromMsg,strlen(nFromMsg),0);

	/* If we're O.K, server should respond with a 250 */
	/* Set the timeout alarm */
	__SMTP_ALARM_SET

	nBytesRecv=recv(hGlobalSocket,(void*)&nRecvBuf,3,0);
	if(nBytesRecv<3)
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : Recieved less than 3 
bytes\n");
		#endif	/* __SMTP_DEBUG_ */

	/* Clear the timeout alarm & see if we timedout */
	__SMTP_ALARM_CLEAR
	if(nSmtpTimedOut==1)
		return(-ESMTPTIMEDOUT);

	nRecvBuf[3]='\0';		/* Terminate after the first 3 bytes, which 
will be just the return code, as we use strcmp on the buffer */

	if(strcmp(nRecvBuf,"250")!=0)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : MAIL FROM: failed\n");
			printf("libsmtp : smtp_sendmail : Server responded 
%s\n",nRecvBuf);
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPINVALIDFROM);
	}

	/* Clear out the rest of the buffer */

	smtp_clear_buf(hGlobalSocket);

	/* Now send the RCPT TO: */
	/* Build our "RCPT TO:" line */
	nLength=strlen(zTo);
	nLength+=12;	/* 12 is the length of the string "RCPT TO: " + \r\n\0 */

	nRcptMsg=(char*)malloc(nLength);

	nRcptMsg=strcpy(nRcptMsg,"RCPT TO: ");
	nRcptMsg=strcat(nRcptMsg,zTo);
	nRcptMsg=strcat(nRcptMsg,"\r\n\0");

	/* Send the RCPT TO: line */
	nBytesSend=send(hGlobalSocket,nRcptMsg,strlen(nRcptMsg),0);

	/* If we're O.K, server should respond with a 250 */

	/* Set the timeout alarm */
	__SMTP_ALARM_SET

	nBytesRecv=recv(hGlobalSocket,(void*)&nRecvBuf,3,0);
	if(nBytesRecv<3)
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : Recieved less than 3 
bytes\n");
		#endif	/* __SMTP_DEBUG_ */

	/* Clear the timeout alarm & see if we timedout */
	__SMTP_ALARM_CLEAR
	if(nSmtpTimedOut==1)
		return(-ESMTPTIMEDOUT);

	nRecvBuf[3]='\0';		/* Terminate after the first 3 bytes, which 
will be just the return code, as we use strcmp on the buffer */

	if(strcmp(nRecvBuf,"250")!=0)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : RCPT TO: failed\n");
			printf("libsmtp : smtp_sendmail : Server responded 
%s\n",nRecvBuf);
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPINVALIDTO);
	}

	/* Clear out the rest of the buffer */

	smtp_clear_buf(hGlobalSocket);

	/* Now send DATA & the message */
	nBytesSend=send(hGlobalSocket,"DATA\r\n",6,0);

	/* DATA should respond with a 354*/
	/* Set the timeout alarm */
	__SMTP_ALARM_SET

	nBytesRecv=recv(hGlobalSocket,(void*)&nRecvBuf,3,0);
	if(nBytesRecv<3)
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : Recieved less than 3 
bytes\n");
		#endif	/* __SMTP_DEBUG_ */

	/* Clear the timeout alarm & see if we timedout */
	__SMTP_ALARM_CLEAR
	if(nSmtpTimedOut==1)
		return(-ESMTPTIMEDOUT);

	nRecvBuf[3]='\0';		/* Terminate after the first 3 bytes, which 
will be just the return code, as we use strcmp on the buffer */

	if(strcmp(nRecvBuf,"354")!=0)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : DATA failed\n");
			printf("libsmtp : smtp_sendmail : Server responded 
%s\n",nRecvBuf);
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPSENDFAILED);
	}

	/* Clear out the rest of the buffer */

	smtp_clear_buf(hGlobalSocket);

	/* Send the message */
	nBytesSend=send(hGlobalSocket,zBody,strlen(zBody),0);

	if(nBytesSend<strlen(zBody))
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : Failed to send entire 
message body\n");
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPSENDFAILED);
	}

	/* Terminate the message with the .CRLF string */
	nBytesSend=send(hGlobalSocket,".\r\n",3,0);

	/* Check the return code to ensure the message was sent O.K */
	/* Set the timeout alarm */
	__SMTP_ALARM_SET

	nBytesRecv=recv(hGlobalSocket,(void*)&nRecvBuf,3,0);
	if(nBytesRecv<3)
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : Recieved less than 3 
bytes\n");
		#endif	/* __SMTP_DEBUG_ */

	/* Clear the timeout alarm & see if we timedout */
	__SMTP_ALARM_CLEAR
	if(nSmtpTimedOut==1)
		return(-ESMTPTIMEDOUT);

	nRecvBuf[3]='\0';		/* Terminate after the first 3 bytes, which 
will be just the return code, as we use strcmp on the buffer */

	if(strcmp(nRecvBuf,"250")!=0)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_sendmail : Send failed after message 
send completed\n");
			printf("libsmtp : smtp_sendmail : Server responded 
%s\n",nRecvBuf);
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPSENDFAILED);
	}

	/* Clear out the rest of the buffer */

	smtp_clear_buf(hGlobalSocket);

	/* All done, message sent & server ready for the next message */
	return(EOK);
}

int smtp_disconnect(void)
{
	/* Disconnect politly from the mail server */

	int nBytesRecv,nBytesSend;	/* Returns from send() & recv() */
	char nRecvBuf[3];				/* Return code recieve buffer 
*/

	/* Send the QUIT command */
	nBytesSend=send(hGlobalSocket,"QUIT\r\n",6,0);

	/* Check to make sure we've disconnected O.K */
	/* Set the timeout alarm */
	__SMTP_ALARM_SET

	nBytesRecv=recv(hGlobalSocket,(void*)&nRecvBuf,3,0);
	if(nBytesRecv<3)
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_disconnect : Recieved less than 3 
bytes\n");
		#endif	/* __SMTP_DEBUG_ */

	/* Clear the timeout alarm & see if we timedout */
	__SMTP_ALARM_CLEAR
	if(nSmtpTimedOut==1)
		return(-ESMTPTIMEDOUT);

	nRecvBuf[3]='\0';		/* Terminate after the first 3 bytes, which 
will be just the return code, as we use strcmp on the buffer */

	if(strcmp(nRecvBuf,"221")!=0)
	{
		#ifdef __SMTP_DEBUG_
			printf("libsmtp : smtp_disconnect : QUIT failed\n");
			printf("libsmtp : smtp_disconnect : Server responded 
%s\n",nRecvBuf);
		#endif	/* __SMTP_DEBUG_ */
		return(-ESMTPDISCONNECTFAILED);
	}

	/* Clear out the rest of the buffer */

	smtp_clear_buf(hGlobalSocket);

	/* Close the socket! */

	close(hGlobalSocket);

	/* Done */
	return(EOK);
}

int smtp_set_timeout(unsigned long nTimeout)
/* Set the new server timeout value */
{
	nSmtpTimeout=nTimeout;
	return(EOK);
}

int smtp_clear_buf(int nSocket)
/* Read through the buffer one byte at a time until we hit a CRLF */
{
	char nBufByte;

	/* Set the timeout alarm */
	__SMTP_ALARM_SET

	while(recv(nSocket,(void*)&nBufByte,1,0)>0)
	{
		if(nBufByte=='\r')
		/* Chomp the \n */
		{
			recv(nSocket,(void*)&nBufByte,1,0);

			__SMTP_ALARM_CLEAR
			if(nSmtpTimedOut==1)	/* See if we timed out */
				break;

			return(EOK);
		}
	}

	return(-1);
}

void smtp_alarm(void)
/* Callback function for alarm() */
{
	#ifdef __SMTP_DEBUG_
		printf("libsmtp : smtp_alarm : Timed out!\n");
	#endif;

	nSmtpTimedOut=1;
	return;
}



